/*============================================================================*/
/**
 * la_interface.c
 * Version: 1.0 (22.11.2010)
 **
/*============================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ftdi/WinTypes.h"
#include "ftdi/ftd2xx.h"

#include "la_interface.h"

/* Empty request package */
static unsigned char request[] = {
	0xFE, 0xFE, 0xFE, 0xFE,
	0xFE, 0xFE, 0xFE, 0xFE,
	0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
};

/* Empty response package */
static unsigned char response[] = {
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
};

/* Flag to prevent a second request to the device */
int second_call_prevent = 0;

/* Last error code */
int last_errorcode = 0;

/* Last error message */
char last_error[2048];



/*============================================================================*/
/*
 * LA_OpenModule
 */
/*============================================================================*/
LA_Handle LA_OpenModule()
{
	unsigned long bytesRead = 0;
	unsigned char buffer[3];
	FT_STATUS ftStatus;
	FT_HANDLE ftHandle;
	struct LA_Module *module;
	LA_Handle handle;

#ifdef DEBUG
	printf("Connecting to usb device DT000003B...\n");
#endif

	/* Open FTDI connection */
	ftStatus = FT_OpenEx("DT000003B", FT_OPEN_BY_SERIAL_NUMBER, &ftHandle);
	if (ftStatus != FT_OK)
	{
		LA_SetError(LA_ERR_DEVICE_NOT_FOUND, 0, "Device not found", __FILE__, __LINE__);
		return NULL;
	}

#ifdef DEBUG
	printf("USB connection established (ftHandle = %p)\n", ftHandle);
#endif

	ftStatus = FT_Read(ftHandle, buffer, 3, &bytesRead);
	if (ftStatus != FT_OK || bytesRead != 3)
	{
		LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "No welcome message received", __FILE__, __LINE__);
		return NULL;
	}

	/* Allocate and initialize module */
	module = malloc(sizeof(struct LA_Module));

	if (module == NULL)
	{
		LA_SetError(LA_ERR_OUT_OF_MEMORY, 0, "Not enough memory for module", __FILE__, __LINE__);
		return NULL;
	}

#ifdef DEBUG
	printf("Module allocation successful (handle = %p)\n", module);
#endif

	module->ftHandle = ftHandle;

	return (LA_Handle) module;
}

/*============================================================================*/
/*
 * LA_CloseModule
 */
/*============================================================================*/
int LA_CloseModule(LA_Handle handle)
{
	if (LA_CheckHandle(handle) != 0)
	{
		return -1;
	}

	/* Close FTDI connection */
	FT_Close(handle->ftHandle);

	/* Free allocated memory */
	free(handle);
	
	return 0;
}

/*============================================================================*/
/*
 * LA_CheckHandle
 */
/*============================================================================*/
int LA_CheckHandle(LA_Handle handle)
{
	if (handle == NULL)
	{
		LA_SetError(LA_ERR_ILLEGAL_HANDLE, 0, "Invalid device handle", __FILE__, __LINE__);
		return -1;
	}

	if (handle->ftHandle == NULL)
	{
		LA_SetError(LA_ERR_ILLEGAL_COM_HANDLE, 0, "Invalid FTDI handle", __FILE__, __LINE__);
		return -1;
	}

	return 0;
}

/*============================================================================*/
/*
 * LA_FTDISendCommand
 */
/*============================================================================*/
int LA_FTDISendCommand(LA_Handle handle, unsigned char command, unsigned long address,
		unsigned char be, unsigned char *data, unsigned long dataLength)
{
	int i;
	unsigned long bytesToWrite, bytesWritten;
	unsigned long bytesToRead, bytesRead;
	FT_STATUS ftStatus;

	if (LA_CheckHandle(handle) != 0)
	{
		return -1;
	}

	if (second_call_prevent != 0) 
	{
		LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "Device busy", __FILE__, __LINE__);
		return -1;
	}
	second_call_prevent++;

	handle->jobID++;

	/* Update request data */
	request[9]	= handle->jobID;
	request[10]	= command;
	request[11]	= (address >>  0) & 0xFF;
	request[12]	= (address >>  8) & 0xFF;
	request[13]	= (address >> 16) & 0xFF;
	request[14]	= (address >> 24) & 0xFF;
	request[15]	= be;
	
	switch (command)
	{
	case CMD_PING:
		bytesToWrite = 16;
		bytesToRead = 7;
		break;

	case CMD_SEND_8_BYTE:
		for (i = 0; i < 8; i++)
		{
			request[16 + i] = (i < dataLength) ? data[i] : 0x00;
		}

		bytesToWrite = 24;
		bytesToRead = 3;
		break;

	case CMD_GET_8_BYTE:
		bytesToWrite = 16;
		bytesToRead = 11;
		break;

	default:
		return -1;
	}

#ifdef DEBUG
	printf("FT_Write Request (%lu): ", bytesToWrite);
	for (i = 0; i < bytesToWrite; i++) printf("%02X", request[i]);
	printf("\n");
#endif

	/* Send request to the device */
	ftStatus = FT_Write(handle->ftHandle, request, bytesToWrite, &bytesWritten);
	if (ftStatus != FT_OK || bytesWritten != bytesToWrite)
	{
		LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "Writing to device failed", __FILE__, __LINE__);
		second_call_prevent--;
		return -1;
	}

	/* Read response from the device */
	ftStatus = FT_Read(handle->ftHandle, response, bytesToRead, &bytesRead);
	if (ftStatus != FT_OK || bytesRead != bytesToRead)
	{
		LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "Reading from device failed", __FILE__, __LINE__);
		second_call_prevent--;
		return -1;
	}

#ifdef DEBUG
	printf("FT_Read Response (%lu): ", bytesRead);
	for (i = 0; i < bytesRead; i++) printf("%02X", response[i]);
	printf("\n");
#endif

	/* Check header, jobID and trailer */
	if (response[0] != 0x1A || response[1] != handle->jobID || response[bytesToRead - 1] != 0x4F)
	{
		LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "Invalid response package format", __FILE__, __LINE__);
		second_call_prevent--;
		return -1;
	}

	memcpy(data, response + 2, dataLength);

	second_call_prevent--;
	
	return 0;
}

/*============================================================================*/
/*
 * LA_FTDISendCommandMultipleBytes
 */
/*============================================================================*/
int LA_FTDISendCommandMultipleBytes(LA_Handle handle, unsigned char command, unsigned long address,
		unsigned long addressDepth, unsigned long iterations, unsigned char *data, unsigned long dataLength)
{
	int i, j;
	unsigned long bufferSize;
	FT_STATUS ftStatus;
	unsigned char *buffer;
	unsigned long bytesToWrite, bytesWritten;
	unsigned long bytesToRead, bytesRead;

#ifdef DEBUG
	printf("LA_FTDISendCommandMultipleBytes(%p, %d, %08lX, %lu, %lu, %p, %lu)\n", handle, command, address, addressDepth, iterations, data, dataLength);
#endif


	if (LA_CheckHandle(handle) != 0)
	{
		return -1;
	}

	if (second_call_prevent != 0) 
	{
		LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "Device busy", __FILE__, __LINE__);
		return -1;
	}
	second_call_prevent++;

	handle->jobID++;

	/* Allocate needed buffer size */
	bufferSize = (addressDepth + 1) * iterations + 20;
	buffer = (unsigned char *)malloc(bufferSize);

	if (buffer == NULL)
	{
		LA_SetError(LA_ERR_OUT_OF_MEMORY, 0, "Not enough memory for buffer", __FILE__, __LINE__);
		second_call_prevent--;
		return -1;
	}


	/* Copy request header into the buffer */
	memcpy(buffer, request, 9);

	/* Update request data */
	buffer[9]	= handle->jobID;
	buffer[10]	= command;
	buffer[11]	= (address >>  0) & 0xFF;
	buffer[12]	= (address >>  8) & 0xFF;
	buffer[13]	= (address >> 16) & 0xFF;
	buffer[14]	= (address >> 24) & 0xFF;
	buffer[15]	= 0x00;

	buffer[16]	= (unsigned char) ((addressDepth >> 0) & 0xFF);
	buffer[17]	= (unsigned char) ((addressDepth >> 8) & 0xFF);
	buffer[18]	= (unsigned char) ((iterations >> 0) & 0xFF);
	buffer[19]	= (unsigned char) ((iterations >> 8) & 0xFF);

	if (command == CMD_SEND_N_BYTE)
	{
		memcpy(buffer + 20, data, dataLength);
		memset(buffer + 20 + dataLength, 0, bufferSize - 20 - dataLength);

		bytesToWrite = dataLength + 20;
		bytesToRead = 5;
	}
	else if (command == CMD_GET_N_BYTE)
	{
		/* Enough memory for response? */
		if (dataLength < iterations * addressDepth)
		{
			LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "Not enough memory for response package provided", __FILE__, __LINE__);
			second_call_prevent--;
			free(buffer);
			return -1;
		}

		bytesToWrite = 20;
		bytesToRead = bufferSize - 20 + 3;
	}

#ifdef DEBUG
	printf("FT_Write Request (%lu): ", bytesToWrite);
	for (i = 0; i < 19; i++) printf("%02X", buffer[i]);
	printf("\n ");
	for (i = 20; i < bytesToWrite; i++)
	{
		printf("%02X", buffer[i]);
		if ((i - 19) % 32 == 0) printf("\n");
		if ((i - 19) % 4 == 0) printf(" ");
	}
	printf("\n");
#endif
	
	/* Submit request */
	ftStatus = FT_Write(handle->ftHandle, buffer, bytesToWrite, &bytesWritten);
	if (ftStatus != FT_OK || bytesWritten != bytesToWrite)
	{
		LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "Writing to device failed", __FILE__, __LINE__);
		second_call_prevent--;
		free(buffer);
		return -1;
	}

	/* Receicve response */
	ftStatus = FT_Read(handle->ftHandle, buffer, bytesToRead, &bytesRead);
	if (ftStatus != FT_OK || bytesRead != bytesToRead)
	{
		LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "Reading from device failed", __FILE__, __LINE__);
		second_call_prevent--;
		free(buffer);
		return -1;
	}

#ifdef DEBUG
	printf("FT_Read Response (%lu):\n ", bytesRead);
	for (i = 0; i < bytesRead; i++)
	{
		printf("%02X", buffer[i]);
		if ((i + 1) % 32 == 0) printf("\n");
		if ((i + 1) % 4 == 0) printf(" ");
	}
	printf("\n");
#endif

	/* Check header, jobID and trailer */
	if (buffer[0] != 0x1A || buffer[1] != handle->jobID || buffer[bytesToRead - 1] != 0x4F)
	{
		LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "Invalid response package format", __FILE__, __LINE__);
		second_call_prevent--;
		free(buffer);
		return -1;
	}

	if (command == CMD_GET_N_BYTE)
	{
		for (i = 0; i < iterations; i++)
		{
			/* Check iteration ID */
			if (buffer[2 + i * addressDepth] != i)
			{
				LA_SetError(LA_ERR_COMMUNICATION_ERROR, 0, "Invalid iteration id", __FILE__, __LINE__);
				second_call_prevent--;
				free(buffer);
				return -1;
			}

			memcpy(data + i * addressDepth, buffer + 2 + i * addressDepth + 1, addressDepth);
		}
	}
	else if (command == CMD_SEND_N_BYTE)
	{
		if (buffer[2] != (iterations & 0xFF) || buffer[3] != ((iterations >> 8) & 0xFF))
		{
			LA_SetError(LA_ERR_COMMUNICATION_ERROR, 1, "Invalid response package format", __FILE__, __LINE__);
			second_call_prevent--;
			free(buffer);
			return -1;
		}
	}

	/* Free allocated buffer */
	free(buffer);

	second_call_prevent--;

	return 0;
}

/*============================================================================*/
/*
 * LA_Ping
 */
/*============================================================================*/
unsigned long LA_Ping(LA_Handle handle, unsigned long value)
{
	unsigned long ping;

	if (LA_FTDISendCommand(handle, CMD_PING, value, 0, (unsigned char *)(&ping), 4) != 0)
	{
		return -1;
	}
	
	return ping & 0xFFFFFFFF;
}

/*============================================================================*/
/*
 * LA_WriteByte
 */
/*============================================================================*/
void LA_WriteByte(LA_Handle handle, unsigned long address, unsigned long value)
{
	unsigned char be;
	unsigned char buff[8];

	be = 1 << (address & 7);

	buff[(address & 7) + 0] = (unsigned char) ((value >>  0) & 0xFF);
	
	LA_FTDISendCommand(handle, CMD_SEND_8_BYTE, address, be, buff, 8);
}
	
/*============================================================================*/
/*
 * LA_WriteWord
 */
/*============================================================================*/
void LA_WriteWord(LA_Handle handle, unsigned long address, unsigned long value)
{
	unsigned char be;
	unsigned char buff[8];

	be = 3 << (address & 6);

	buff[(address & 6) + 0] = (unsigned char) ((value >>  0) & 0xFF);
	buff[(address & 6) + 1] = (unsigned char) ((value >>  8) & 0xFF);
	
	LA_FTDISendCommand(handle, CMD_SEND_8_BYTE, address, be, buff, 8);
}
	
/*============================================================================*/
/*
 * LA_WriteLong
 */
/*============================================================================*/
void LA_WriteLong(LA_Handle handle, unsigned long address, unsigned long value)
{
	unsigned char be;
	unsigned char buff[8];

	be = 15 << (address & 4);

	buff[(address & 4) + 0] = (unsigned char) ((value >>  0) & 0xFF);
	buff[(address & 4) + 1] = (unsigned char) ((value >>  8) & 0xFF);
	buff[(address & 4) + 2] = (unsigned char) ((value >> 16) & 0xFF);
	buff[(address & 4) + 3] = (unsigned char) ((value >> 24) & 0xFF);
	
	LA_FTDISendCommand(handle, CMD_SEND_8_BYTE, address, be, buff, 8);
}
		
/*============================================================================*/
/*
 * LA_WriteLongLong
 */
/*============================================================================*/
void LA_WriteLongLong(LA_Handle handle, unsigned long address, unsigned long long value)
{
	LA_FTDISendCommand(handle, CMD_SEND_8_BYTE, address, 0xFF, (unsigned char *)(&value), 8);
}
		
/*============================================================================*/
/*
 * LA_ReadByte
 */
/*============================================================================*/
unsigned long LA_ReadByte(LA_Handle handle, unsigned long address)
{
	unsigned char be;
	unsigned char buff[8];

	be = 1 << (address & 7);

	if (LA_FTDISendCommand(handle, CMD_GET_8_BYTE, address, be, buff, 8) != 0)
	{
		return 0;
	}

	return *((unsigned long *)(buff + (address & 7))) & 0xFF;
}
		
/*============================================================================*/
/*
 * LA_ReadWord
 */
/*============================================================================*/
unsigned long LA_ReadWord(LA_Handle handle, unsigned long address)
{
	unsigned char be;
	unsigned char buff[8];

	be = 3 << (address & 6);

	if (LA_FTDISendCommand(handle, CMD_GET_8_BYTE, address, be, buff, 8) != 0)
	{
		return 0;
	}

	return *((unsigned long *)(buff + (address & 6))) & 0xFFFF;
}
		
/*============================================================================*/
/*
 * LA_ReadLong
 */
/*============================================================================*/
unsigned long LA_ReadLong(LA_Handle handle, unsigned long address)
{
	unsigned char be;
	unsigned long data;
	unsigned char buff[8];

	be = 15 << (address & 4);

	if (LA_FTDISendCommand(handle, CMD_GET_8_BYTE, address, be, buff, 8) != 0)
	{
		return 0;
	}

	return *((unsigned long *)(buff + (address & 4))) & 0xFFFFFFFF;
}
		
/*============================================================================*/
/*
 * LA_ReadLongLong
 */
/*============================================================================*/
unsigned long long LA_ReadLongLong(LA_Handle handle, unsigned long address)
{
	unsigned long long data;

	if (LA_FTDISendCommand(handle, CMD_GET_8_BYTE, address, 0xFF, (unsigned char *)(&data), 8) != 0)
	{
		return 0;
	}

	return data;
}
		
/*============================================================================*/
/*
 * LA_ReadMultipleBytes
 */
/*============================================================================*/
int LA_ReadMultipleBytes(LA_Handle handle, unsigned long address, unsigned long addressDepth,
	unsigned long iterations, unsigned char *buffer, unsigned long bufferLength)
{
#ifdef DEBUG
	printf("LA_ReadMultipleBytes(%p, %08lX, %lu, %lu, %p, %lu)\n", handle, address, addressDepth, iterations, buffer, bufferLength);
#endif

	/* Buffer size large enough? */
	if (bufferLength < iterations * addressDepth)
	{
		LA_SetError(LA_ERR_ILLEGAL_ARGUMENTS, 0, "Invalid buffer size", __FILE__, __LINE__);
		return -1;
	}

	while (iterations > 32768)
	{
		if (LA_FTDISendCommandMultipleBytes(handle, CMD_GET_N_BYTE, address, addressDepth, 32768, buffer, 32768 * addressDepth) != 0)
		{
			return -1;
		}

		iterations -= 32768;
		buffer += 32768 * addressDepth;
		bufferLength -= 32768 * addressDepth;
	}

	return LA_FTDISendCommandMultipleBytes(handle, CMD_GET_N_BYTE, address, addressDepth, iterations, buffer, bufferLength);
}

/*============================================================================*/
/*
 * LA_WriteMultipleBytes
 */
/*============================================================================*/
int LA_WriteMultipleBytes(LA_Handle handle, unsigned long address, unsigned long addressDepth,
		unsigned long iterations, unsigned char *buffer, unsigned long bufferLength)
{
#ifdef DEBUG
	printf("LA_WriteMultipleBytes(%p, %08lX, %lu, %lu, %p, %lu)\n", handle, address, addressDepth, iterations, buffer, bufferLength);
#endif

	/* Buffer size large enough? */
	if (bufferLength < iterations * addressDepth)
	{
		LA_SetError(LA_ERR_ILLEGAL_ARGUMENTS, 0, "Invalid buffer size", __FILE__, __LINE__);
		return -1;
	}

	while (iterations > 32768)
	{
		if (LA_FTDISendCommandMultipleBytes(handle, CMD_SEND_N_BYTE, address, addressDepth, 32768, buffer, 32768 * addressDepth) != 0)
		{
			return -1;
		}

		iterations -= 32768;
		buffer += 32768 * addressDepth;
		bufferLength -= 32768 * addressDepth;
	}

	return LA_FTDISendCommandMultipleBytes(handle, CMD_SEND_N_BYTE, address, addressDepth, iterations, buffer, bufferLength);
}

/*============================================================================*/
/*
 * LA_SetError
 */
/*============================================================================*/
void LA_SetError(int errorcode, int suberror, const char *errortext, const char *filename, unsigned line)
{
	last_errorcode = errorcode;
	sprintf(last_error, "%s (%d.%d) in file %s line %d\n", errortext, errorcode, suberror, filename, line);

#ifdef DEBUG
	printf("Error: %s", last_error);
#endif
}

/*============================================================================*/
/*
 * LA_ClearLastError
 */
/*============================================================================*/
void LA_ClearLastError()
{
	last_errorcode = 0;
}

/*============================================================================*/
/*
 * LA_GetLastErrorText
 */
/*============================================================================*/
const char *LA_GetLastErrorText()
{
	if (last_errorcode == 0)
	{
		return 0;
	}

	return last_error;
}

/*============================================================================*/
/*
 * LA_GetLastError
 */
/*============================================================================*/
int LA_GetLastError()
{
	return last_errorcode;
}


