
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "la/la_interface.h"

#include "sampler.h"


/* Globales Handle auf den USB-LOGI */
LA_Handle la_handle = NULL;

/* Debug Readback der Register des LA */
static void printStatus();

/* Interne Hilfsfunktionen zum Setzen von Bits in Registern */
void set_bit(unsigned long address, unsigned char bit)
{
	unsigned long value;

	value = LA_ReadLong(la_handle, address);
	value |= (1 << bit);
	LA_WriteLong(la_handle, address, value);
}

void clear_bit(unsigned long address, unsigned char bit)
{
	unsigned long value;

	value = LA_ReadLong(la_handle, address);
	value &= ~(1 << bit);
	LA_WriteLong(la_handle, address, value);
}


/*******************************************************************************
Zum USB-LOGI Verbinden und die Verbindung testen

Rückgabewerte:
 0 = Alles OK
-1 = Konnte keine Verbindung aufbauen
-2 = Ping fehlgeschlagen
********************************************************************************/
int la_connect()
{
	/* Verbinden wenn keine Verbindung besteht */
	if (la_handle == NULL)
	{
		la_handle = LA_OpenModule();

		/* Prüfen ob Verbindung geklappt hat */
		if (la_handle == NULL)
		{
			return -1;
		}

		/* Ping senden um Verbindung zu testen */
		if (LA_Ping(la_handle, 0xDEADBEEF) != 0xDEADBEEF)
		{
			return -2;
		}
	}

	return 0;
}

/*******************************************************************************
Verbindung zum USB-LOGI trennen

Rückgabewerte:
 0 = Alles OK
-1 = Keine bestehende Verbindung
********************************************************************************/
int la_disconnect()
{
	if (LA_CheckHandle(la_handle) != 0)
	{
		return -1;
	}

	LA_CloseModule(la_handle);
	la_handle = NULL;

	return 0;
}

/*******************************************************************************
Ein Array von 36 Triggern wird in ein Array mit 4 unsigned long long Variablen
umgewandelt.

********************************************************************************/
void la_build_trigger_reg_values(struct la_trigger *trigger, unsigned long long *reg_values)
{
	int i;

	reg_values[0] = 0;
	reg_values[1] = 0;
	reg_values[2] = 0;
	reg_values[3] = 0;

	for (i = 0; i < LA_NUM_CHANNELS; i++)
	{
		reg_values[0] |= trigger[i].low << i;
		reg_values[1] |= trigger[i].high << i;
		reg_values[2] |= trigger[i].rising_edge << i;
		reg_values[3] |= trigger[i].falling_edge << i;
	}
}

/*******************************************************************************
USB-LOGI konfigurieren (Trigger, Sample Rate, etc...)

Rückgabewerte:
 0 = Alles OK
-1 = Keine bestehende Verbindung
-2 = Beschreiben der Register fehlgeschlagen
********************************************************************************/
int la_configure(struct la_config *cfg)
{
	unsigned long cfg0_l_value, cfg0_h_value;
	unsigned long long triggerA[4], triggerB[4];

	if (LA_CheckHandle(la_handle) != 0)
	{
		return -1;
	}

	cfg0_l_value = 0;
	cfg0_h_value = 0;

	/* Trigger Ignore? */
	cfg0_l_value |= (cfg->trigger_ignore) ? (1 << LA_BIT_TRIGGER_IGNORE) : 0;
	
	/* Pretrigger setzen */
	cfg0_l_value |= ((cfg->pretrigger & 7) << LA_BIT_PRETRIGGER);

	/* Sample Rate einstellen */
	if (cfg->sample_rate >= 500000000)
	{
		/* 500 MSamples pro Sekunde */
		cfg0_l_value |= (1 << LA_BIT_MODE_500MS);
	}
	else if (cfg->sample_rate == 250000000)
	{
		/* 250 MSamples pro Sekunde */
		cfg0_l_value |= (1 << LA_BIT_MODE_250MS);
	}
	else
	{
		/* Bei anderen Abtastraten muss ein Wert zum Teilen
		   des 150 MHz Basistaktes eingestellt werden */
		unsigned long div;

		div = (cfg->sample_rate != 0) ? (LA_SAMPLE_CLK / cfg->sample_rate) - 1 : 0;
		cfg0_h_value |= ((div & 0xFFFFFF) << LA_BIT_CLK_DIVIDER);
	}

	/* Konfigurationsregister beschreiben */
	LA_WriteLong(la_handle, LA_ADDR_CFG0_L, cfg0_l_value);
	LA_WriteLong(la_handle, LA_ADDR_CFG0_H, cfg0_h_value);

	if (LA_GetLastError() != 0)
	{
		return -2;
	}

	/* Trigger schreiben */
	la_build_trigger_reg_values(cfg->triggerA, triggerA);
	la_build_trigger_reg_values(cfg->triggerB, triggerB);

	LA_WriteLongLong(la_handle, LA_ADDR_TRIGGER_A_0, triggerA[0]);
	LA_WriteLongLong(la_handle, LA_ADDR_TRIGGER_A_1, triggerA[1]);
	LA_WriteLongLong(la_handle, LA_ADDR_TRIGGER_A_RE, triggerA[2]);
	LA_WriteLongLong(la_handle, LA_ADDR_TRIGGER_A_FE, triggerA[3]);

	LA_WriteLongLong(la_handle, LA_ADDR_TRIGGER_B_0, triggerB[0]);
	LA_WriteLongLong(la_handle, LA_ADDR_TRIGGER_B_1, triggerB[1]);
	LA_WriteLongLong(la_handle, LA_ADDR_TRIGGER_B_RE, triggerB[2]);
	LA_WriteLongLong(la_handle, LA_ADDR_TRIGGER_B_FE, triggerB[3]);

	LA_WriteByte(la_handle, LA_ADDR_TRIGGER_HELPER_0, cfg->trigger_helper0);
	LA_WriteByte(la_handle, LA_ADDR_TRIGGER_HELPER_1, cfg->trigger_helper1);
	LA_WriteByte(la_handle, LA_ADDR_TRIGGER_SELECTION, cfg->trigger_selection);
	LA_WriteByte(la_handle, LA_ADDR_TRIGGER_START, cfg->trigger_start_trigger);
	LA_WriteWord(la_handle, LA_ADDR_TRIGGER_COUNTER_0, cfg->trigger_counter);

	return 0;
}

/*******************************************************************************
Aufzeichnung starten

Rückgabewerte:
 0 = Alles OK
-1 = Keine bestehende Verbindung
-2 = Beschreiben der Register fehlgeschlagen
********************************************************************************/
int la_start_sampling()
{
	if (LA_CheckHandle(la_handle) != 0)
	{
		return -1;
	}

	/* Zum Starten der Aufzeichnung muss das Startbit kurz auf 1 gesetzt werden */
	set_bit(LA_ADDR_CFG0_L, LA_BIT_START_SAMPLING);
	clear_bit(LA_ADDR_CFG0_L, LA_BIT_START_SAMPLING);

	if (LA_GetLastError() != 0)
	{
		return -2;
	}

	return 0;
}

/*******************************************************************************
Aufzeichnung stoppen

Rückgabewerte:
 0 = Alles OK
-1 = Keine bestehende Verbindung
-2 = Beschreiben der Register fehlgeschlagen
********************************************************************************/
int la_stop_sampling()
{
	if (LA_CheckHandle(la_handle) != 0)
	{
		return -1;
	}

	/* Zum Stoppen der Aufzeichnung muss das Stoppbit kurz auf 1 gesetzt werden */
	set_bit(LA_ADDR_CFG0_L, LA_BIT_STOP_SAMPLING);
	clear_bit(LA_ADDR_CFG0_L, LA_BIT_STOP_SAMPLING);

	if (LA_GetLastError() != 0)
	{
		return -2;
	}

	return 0;
}

/*******************************************************************************
Auf Fehler testen und Fehlermeldung ausgeben falls nötig.
********************************************************************************/
void la_test_error(int do_exit)
{
	if (LA_GetLastError() != 0)
	{
		fprintf(stderr, "Fehler aufgetreten:\n%s\n", LA_GetLastErrorText());
		if (do_exit)
		{
			exit(0);
		}
	}
}

/*******************************************************************************
Aktuellen Wert des Adress-Zählers abfragen.

Rückgabewerte:
>= 0 = Aktueller Wert des Adress Zählers.
-1 = Keine bestehende Verbindung
********************************************************************************/
unsigned long la_get_addr_cnt()
{
	if (LA_CheckHandle(la_handle) != 0)
	{
		return -1;
	}

	return (LA_ReadLong(la_handle, LA_ADDR_ADDR_CNT) << 2) & LA_RAM_ADDR_MASK;
}

/*******************************************************************************
Aktuellen Wert des Trigger-Zählers abfragen.

Rückgabewerte:
>= 0 = Aktueller Wert des Trigger Zählers.
-1 = Keine bestehende Verbindung
********************************************************************************/
unsigned long la_get_trigger_cnt()
{
	if (LA_CheckHandle(la_handle) != 0)
	{
		return -1;
	}

	return LA_ReadWord(la_handle, LA_ADDR_TRIGGER_CNT);
}

/*******************************************************************************
Messdaten vom USB-LOGI holen.

Rückgabewerte:
 0 = Alles OK
-1 = Keine bestehende Verbindung
-2 = Buffer zu klein
********************************************************************************/
int la_get_sampling_data(unsigned long long *data, unsigned long length)
{
	if (LA_CheckHandle(la_handle) != 0)
	{
		return -1;
	}

	if (length < LA_NUM_SAMPLES)
	{
		return -2;
	}

	set_bit(LA_ADDR_CFG0_L, LA_BIT_INTERNAL_RAM);

	/* Vier Speicher a 1024 Samples (= 8*1024 Bytes) auslesen */
	LA_ReadMultipleBytes(la_handle, LA_ADDR_RAM0, 8*1024, 1, (unsigned char *)(data),        8*1024);
	LA_ReadMultipleBytes(la_handle, LA_ADDR_RAM1, 8*1024, 1, (unsigned char *)(data + 1024), 8*1024);
	LA_ReadMultipleBytes(la_handle, LA_ADDR_RAM2, 8*1024, 1, (unsigned char *)(data + 2048), 8*1024);
	LA_ReadMultipleBytes(la_handle, LA_ADDR_RAM3, 8*1024, 1, (unsigned char *)(data + 3072), 8*1024);
	
	clear_bit(LA_ADDR_CFG0_L, LA_BIT_INTERNAL_RAM);

	return 0;
}

/*******************************************************************************
Status der Messung abfragen.
Im Parameter progress wird der Status in Prozent (0 .. 100) gespeichert.
Damit das richtig funktioniert, muss die Funktion am Anfang der Statusabfragen
einmal mit NULL als Wert für progress aufgerufen werden.

Rückgabewerte:
>= 0 = Status des LA (LA_STATUS_* Makros)
-1 = Keine bestehende Verbindung
********************************************************************************/
int la_get_status(int *progress)
{
	int status;
	unsigned long addr;
	static unsigned long first_addr;

	if (LA_CheckHandle(la_handle) != 0)
	{
		return -1;
	}

	/* Status und aktuelle Speicheradresse abfragen */
	status = LA_ReadLong(la_handle, LA_ADDR_STATUS) & 0x0F;
	addr = la_get_addr_cnt();

	if (LA_GetLastError() != 0)
	{
		return -2;
	}

	/* Aktuellen Fortschritt berechnen und in progress speichern */
	if (progress != NULL)
	{
		unsigned long diff = (addr + (LA_RAM_ADDR_MASK + 1) - first_addr) & LA_RAM_ADDR_MASK;

		*progress = (diff * 100) / LA_RAM_ADDR_MASK;
	}
	/* Speicheradresse am Anfang er Abfragen speichern */
	else
	{
		first_addr = addr;
	}

	return status;
}



/*
Debugausgabe der Konfigurations-, Status- und Triggerregister
*/
static void printBin(unsigned long long value)
{
	int i;
	for (i = LA_NUM_CHANNELS-1; i >= 0; i--)
	{
		printf("%c", (value & (1 << i)) ? '1' : '0');
	}
}
static void printStatus()
{
	LA_Handle handle = la_handle;
	printf("Cfg Reg 0: %08lX %08lX\n", LA_ReadLong(handle, 0x0000), LA_ReadLong(handle, 0x0004));
	printf("Cfg Reg 1: %08lX %08lX\n", LA_ReadLong(handle, 0x0008), LA_ReadLong(handle, 0x000C));
	printf("Status:    %08lX\n", LA_ReadLong(handle, 0x00E8));
	printf("Trigger A (0):  "); printBin(LA_ReadLongLong(handle, 0x0010)); printf("\n");
	printf("Trigger A (1):  "); printBin(LA_ReadLongLong(handle, 0x0018)); printf("\n");
	printf("Trigger A (RE): "); printBin(LA_ReadLongLong(handle, 0x0020)); printf("\n");
	printf("Trigger A (FE): "); printBin(LA_ReadLongLong(handle, 0x0028)); printf("\n");
	printf("Trigger B (0):  "); printBin(LA_ReadLongLong(handle, 0x0030)); printf("\n");
	printf("Trigger B (1):  "); printBin(LA_ReadLongLong(handle, 0x0038)); printf("\n");
	printf("Trigger B (RE): "); printBin(LA_ReadLongLong(handle, 0x0040)); printf("\n");
	printf("Trigger B (FE): "); printBin(LA_ReadLongLong(handle, 0x0048)); printf("\n");
	printf("%02lX %02lX %02lX %02lX\n", LA_ReadByte(handle, 0x0050), LA_ReadByte(handle, 0x0051), LA_ReadByte(handle, 0x0052), LA_ReadByte(handle, 0x0053));
	printf("%04lX\n", LA_ReadLong(handle, 0x0058) & 0xFFFF);
}

